/* evtgen engine.
   Copyright (C) 2003 Julien Olivain and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of evtgen.

   evtgen is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   evtgen is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with evtgen; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
** evtgen.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "evtgen.h"

evtgen_t *
new_evtgen_context(void)
{
  evtgen_t *ctx;

  ctx = xzmalloc(sizeof (evtgen_t));
  ctx->max_event = DEFAULT_MAX_EVENT;
  ctx->delay = DEFAULT_RTDELAY;
  ctx->bound = DEFAULT_RTBOUND;
  ctx->builtin_var_hash = new_strhash(1021);
  ctx->randvar_hash = new_strhash(1021);

  register_builtin_vars(ctx);

  return (ctx);
}

/* return a floating random number in [0;1.0[ */
float
frand(void)
{
  return ((rand()) / (RAND_MAX + 1.0f));
}

int
generate_string(evtgen_t *ctx, random_var_t *randvar, char *buffer, size_t s)
{
  value_t *e;
  text_token_t *t;
  int ret;
  int total;
  float rnd;
  float threshold;

  total = 0;
  ret = 0;
  threshold = 0.0f;
  rnd = frand();
  /* chose a random value */
  for (e = randvar->event_list; e; e = e->next)
    {
      threshold += e->prob;
      if (rnd < threshold)
        break;
      else if (e->next == NULL)
        break;
      /* this condition is used to get rid of errors
       * due to float32 approximations */
    }
  /* generate string */
  for (t = e->token_list; t; t = t->generic.next)
    {
      switch (t->generic.type)
        {
        case TOKEN_STRING:
          ret = snprintf(buffer, s, "%s", t->string.str);
          break;

        case TOKEN_BUILTIN_VAR:
          ret = t->builtin.var->expand_func(ctx, buffer, s);
          break;

        case TOKEN_VAR:
          ret = generate_string(ctx, t->randvar.var, buffer, s);
          break;

        default:
          fprintf(stderr, "Error...\n");
          exit(EXIT_FAILURE);
          break;
        }

      if (ret < 0)
        {
          fprintf(stderr, "error\n");
          exit(EXIT_FAILURE);
        }
      buffer += ret;
      s -= ret;
      total += ret;
    }

  return (total);
}

void
fprintf_value_list(FILE *fp, evtgen_t *ctx, text_token_t *parent_tok,
                   random_var_t *var, float p, char *str, int depth)
{
  value_t *e;
  text_token_t *t;
  char base[2048];
  int s, rs = 2048, idx = 0;

  if (var) {
    for (e = var->event_list; e; e = e->next) {
      base[0] = '\0';
      strncpy(base, str, 2048);
      idx = strlen(base);
      rs -= idx;
      for (t = e->token_list; t; t = t->generic.next) {
        if (t->generic.type == TOKEN_STRING) {
          s = snprintf(base + idx, rs, "\"%s\" ", t->string.str);
          rs -= s;
          idx += s;
        }
        else if (t->generic.type == TOKEN_BUILTIN_VAR) {
          s = snprintf(base + idx, rs, "$%s ", t->builtin.var->name);
          rs -= s;
          idx += s;
        }
        else if (t->generic.type == TOKEN_VAR) {
          fprintf_value_list(fp, ctx, t, t->randvar.var, p * e->prob, base, depth + 1);
          break;
        }
        else {
          fprintf(stderr, "Error...\n");
          exit(EXIT_FAILURE);
        }
      }
      if (parent_tok)
        fprintf_value_list(fp, ctx, parent_tok->generic.next, NULL, p * e->prob, base, depth - 1);
    }
  } else {
    base[0] = '\0';
    strncpy(base, str, 2048);
    idx = strlen(base);
    rs -= idx;
    for (t = parent_tok; t; t = t->generic.next) {
      if (t->generic.type == TOKEN_STRING) {
        s = snprintf(base + idx, rs, "\"%s\" ", t->string.str);
        rs -= s;
        idx += s;
      }
      else if (t->generic.type == TOKEN_BUILTIN_VAR) {
        s = snprintf(base + idx, rs, "$%s ", t->builtin.var->name);
        rs -= s;
        idx += s;
      }
      else if (t->generic.type == TOKEN_VAR) {
        fprintf_value_list(fp, ctx, t, t->randvar.var, p, base, depth + 1);
        break;
      }
      else {
        fprintf(stderr, "Error...\n");
        exit(EXIT_FAILURE);
      }
    }
    if (depth == 0)
      fprintf(fp, "p=%12.10f: %s\n", p, base);
  }

  return;
}

static int
expand_seq(evtgen_t *ctx, char *buf, size_t s)
{
  return( snprintf(buf, s, "%lu", ctx->seq) );
}

static int
expand_ctime(evtgen_t *ctx, char *buf, size_t s)
{
  return( snprintf(buf, s, "%lu", ctx->cur_time) );
}

static int
expand_strctime(evtgen_t *ctx, char *buf, size_t s)
{
  return (strftime(buf, s, "%a %b %d %H:%M:%S %Y", localtime(&ctx->cur_time)));
}

static int
expand_rnd16(evtgen_t *ctx, char *buf, size_t s)
{
  /* low bit of the rand()'s LCG are 'less random'... */
  return ( snprintf(buf, s, "%i", rand() & 0xffff) );
}

static builtin_record_t vars_g[] = {
  { "seq", expand_seq, "event sequence number" },
  { "ctime", expand_ctime, "Unix C time (int)" },
  { "strctime", expand_strctime, "Unix string time" },
  { "rnd16", expand_rnd16, "16-bit random number in [0;65536["},
  { NULL, NULL, NULL }
};

void
register_builtin_vars(evtgen_t *ctx)
{
  int v;
  strhash_t *h;

  h = ctx->builtin_var_hash;
  for (v = 0; vars_g[v].name; v++)
    {
      /* printf("%-10s : %s\n", vars_g[v].name, vars_g[v].description); */
      if (strhash_get(h, vars_g[v].name) != NULL)
        {
          fprintf(stderr, "error: builtin var $%s already defined\n",
                  vars_g[v].name);
          exit(EXIT_FAILURE);
        }
      strhash_add(h, &vars_g[v], vars_g[v].name);
    }
}

void
fprintf_event(FILE *fp, value_t *event)
{
  text_token_t *t;

  fprintf(fp, "%8.6f: ", event->prob);
  for (t = event->token_list; t; t = t->generic.next)
    {
      switch (t->generic.type)
        {
        case TOKEN_STRING:
          fprintf(fp, "\"%s\" ", t->string.str);
          break;

        case TOKEN_BUILTIN_VAR:
          fprintf(fp, "$%s ", t->builtin.var->name);
/*           fprintf(fp, "$var "); */
          break;

        case TOKEN_VAR:
          fprintf(fp, "@%s ", t->randvar.name);
          break;

        default:
          fprintf(fp, "unhandled token type\n");
          exit(EXIT_FAILURE);
          break;
        }
    }
  fprintf(fp, "\n");
}

void
fprintf_randvar(FILE *fp, random_var_t *var)
{
  value_t *e;
  int i;

  for (e = var->event_list, i = 0; e; e = e->next, i++)
    {
      fprintf(fp, "value %3i: ", i);
      fprintf_event(fp, e);
    }
}

void
fprintf_randvar_list(FILE *fp, evtgen_t *ctx)
{
  random_var_t *x;

  for (x = ctx->randvar_list; x; x = x->next)
    {
      fprintf(fp, "Random variable @%s:\n", x->name);
      fprintf_randvar(fp, x);
      fprintf(fp, "\n");
    }
}
