/* evtgen headers.
   Copyright (C) 2003 Julien Olivain and LSV, CNRS UMR 8643 & ENS Cachan.

   This file is part of evtgen.

   evtgen is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   evtgen is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with evtgen; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * event-generator.h
 */

#ifndef EVTGEN_H
#define EVTGEN_H

#include <unistd.h>

#include "strhash.h"

#define BUFFERSZ 2048

#define DEFAULT_CONFIG_FILE "evtgen.conf"
#define DEFAULT_MAX_EVENT 100
#define DEFAULT_RTDELAY 250
#define DEFAULT_RTBOUND 0

#define MODE_REALTIME 1
#define MODE_VERBOSE  2

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#define TOKEN_STRING      1
#define TOKEN_BUILTIN_VAR 2
#define TOKEN_VAR         3

typedef struct evtgen_s evtgen_t;
typedef struct generic_token_s generic_token_t;
typedef struct string_token_s string_token_t;
typedef struct builtin_var_token_s builtin_var_token_t;
typedef struct random_var_token_s random_var_token_t;
typedef union text_token_u text_token_t;
typedef struct random_var_s random_var_t;
typedef struct symbol_token_s symbol_token_t;
typedef struct compiler_s compiler_t;

typedef struct builtin_record_s builtin_record_t;

struct evtgen_s
{
  random_var_t *prob_tree_root;
  random_var_t *randvar_list;
  unsigned long seq;
  unsigned long max_event;
  time_t cur_time;
  unsigned long flags;
  strhash_t *builtin_var_hash;
  strhash_t *randvar_hash;
  char buffer[BUFFERSZ];
  int index;
  char *config_file;
  int prng_seed;
  int delay;
  int bound;
};

typedef int (*expand_t)(evtgen_t *ctx, char *buf, size_t s);

struct builtin_record_s
{
  char *name;
  expand_t expand_func;
  char *description;
};

struct symbol_token_s
{
  char *name;  /**< Symbol name */
  int line;    /**< Declaration line */
};

struct generic_token_s
{
  int type;
  text_token_t *next;
};

struct string_token_s
{
  int type;
  text_token_t *next;
  char *str;
};

struct builtin_var_token_s
{
  int type;
  text_token_t *next;
  builtin_record_t *var;
};

struct random_var_token_s
{
  int type;
  text_token_t *next;
  char *name; /* unresolved name (1st pass) */
  random_var_t *var; /* resolved randvar declaration (2nd pass) */
};


union text_token_u
{
  generic_token_t     generic;
  string_token_t      string;
  builtin_var_token_t builtin;
  random_var_token_t  randvar;
};

typedef struct value_s value_t;
struct value_s
{
  float prob;
  text_token_t *token_list;
  value_t *next;
};

struct random_var_s
{
  float allocated_prob;
  int num_unallocated_prob;
  unsigned long flags;
  value_t *event_list;
  int line;
  char *name;
  random_var_t *next;
};

/* --- */

/* evtgen.c */
evtgen_t *new_evtgen_context(void);
void register_builtin_vars(evtgen_t *ctx);
void fprintf_event(FILE *fp, value_t *event);
void fprintf_randvar(FILE *fp, random_var_t *var);
void fprintf_randvar_list(FILE *fp, evtgen_t *ctx);
float frand(void);
void fprintf_value_list(FILE *fp, evtgen_t *ctx, text_token_t *parent_tok, random_var_t *var, float p, char *str, int depth);

/* compiler.c */
void compile_config_file(evtgen_t *ctx);
text_token_t *build_string_token(char *str);
text_token_t *build_builtin_var_token(char *name);
text_token_t *build_randvar_token(char *name);
value_t *build_event(float prob, text_token_t *token_list);
random_var_t *build_randvar(char *name, int line, value_t *event_list);
void build_randvar_list(random_var_t *randvar_list);
int generate_string(evtgen_t *ctx, random_var_t *randvar,
                    char *buffer, size_t s);

/* issdl.l */
int parserlex(void);
void parsererror(char *s);

/* issdl.y */
int parserparse(void);

/* util.c */
void *xzmalloc(size_t size);
void *xmalloc(size_t size);
void *xrealloc(void *ptr, size_t size);
FILE *xfopen(const char *filename, const char *mode);

/*
** Debugging stuffs...
******/

#ifdef EVTGEN_DEBUG

#define DPRINTF(args) \
  do { \
    printf("%24s:%-4d: ", __FILE__, __LINE__); \
    printf args ; \
  } while (0)

#define DFPRINTF(args) fprintf args

#else /* DEBUG */
# define DPRINTF(args)
# define DFPRINTF(args)
#endif /* DEBUG */


#endif /* EVENT_GEN_H */
